import click
from .types import *
import kontiguity.load as kload

@click.command("load")
@click.option(
    '-n',
    '--name', 
    type=str,
    default="",
    help="name of the experiment (recommanded: species name. info: spaces are not allowed and will be replaced by _.)"
)
@click.option(
    '-o',
    '--outpath',
    type=str,
    help="output folder path, created if non-existent"
)
@click.option(
    "-r",
    "--ref",
    type=str,
    help="path to the reference genome fasta OR the GCA reference which will automatically be loaded from ENA database."
)
@click.option(
    "--chroms",
    type=str,
    help='path to a chromosome information file detailing the type of each sequence present in the reference (Mandatory column heads: ["id", "sequence_type", "sequence_name"]). "sequence_type" must be in the ENA database format : ["chromosome", "organelle", ...]. Required only for a local fasta, GCA referenced genomes will have the chromosome.tsv generated.'
)
@click.option(
    "--wgs",
    type=PAIR_LIST,
    help="comma-separated list of paths to the WGS fastq(s) OR SRA accession. If paired and local, provide both fastqs separated by : ."
)
@click.option(
    "--hic",
    type=PAIR_LIST,
    help="comma-separated list of paths to the Hi-C fastq(s) OR SRA accession. If paired and local, provide both fastqs separated by : "
)
@click.option(
    "--table",
    type=str,
    help='path to a csv table providing the data parameters (Mandatory column heads: ["name", "ref", "wgs", "hic"]). See test_data/load/test_dataset.csv for format example.'
)
@click.option(
    "--dtol",
    is_flag=True,
    default=False,
    help="if selected, a data table will be created and loaded from the Darwin Tree of Life project database."
)
@click.option(
    "--no_wgs",
    is_flag=True,
    default=False,
    help="if selected, the WGS data is not loaded nor required. Usefull for calling Logan in contigs reconstruction (see retrieve command)."
)
@click.option(
    "--no_hic",
    is_flag=True,
    default=False,
    help="if selected, the Hi-C data is not loaded nor required. Usefull when using kontiguity for contigs retrieval only."
)
@click.option(
    "-t",
    "--threads",
    type=int,
    default=8,
    help='number of threads to launch for each subtask (dflt: 8)'
)
@click.option(
    "--sbatch",
    is_flag=True,
    default=False,
    help="if selected, all the bash script will be launched as individual jobs on a SLURM distribution."
)
@click.option(
    "--sbtach_partition",
    default='dedicated',
    type=str,
    help="partition requested for sbatch."
)
@click.option(
    "--sbtach_qos",
    default= 'fast',
    type=str,
    help="quality of service required for sbatch."
)
@click.option(
    "--sbtach_mem",
    default='40G',
    type=str,
    help="minimum amount of real memory requested for sbatch."
)
@click.option(
    "--sbatch_ncpus",
    default=30,
    type=int,
    help="number of cpus required per task fro sbatch."
)
def load(**args):
    kload.load(**args)


from kontiguity.utils.imports import *

class DToLScrapperScheduler(threading.Thread):
    def __init__(self, input_queue, output_queue, **args):
        super(DToLScrapperScheduler, self).__init__(**args)
        self._input_queue = input_queue
        self._output_queue = output_queue
        self.start()

    def run(self):
        while True:
            try:
                value = self._input_queue.get(timeout=10)
            except Empty:
                break
            if value == "DONE":
                break
            limit, offset = value
            scrapper = DToLScrapper(limit, offset)
            results = scrapper.scrap()

            for result in results:
                self._output_queue.put(result)
            
class DToLScrapper():
    def __init__(self, limit, offset):
        self.URL = "https://portal.darwintreeoflife.org/api/data_portal"
        self.URL_args = {
            "limit" : limit,
            "offset" : offset,
            "sort" : "currentStatus:asc",
            "current_class" : "kingdom"
        }

    def get_url(self):
        url = self.URL + "?"
        for arg in self.URL_args.keys():
            url += f"{arg}={self.URL_args[arg]}&"
        return url[:-1]
    
    def scrap(self):
        current_url = self.get_url()
        try:
            result = requests.get(current_url)
        except:
            print("Table not received")
        return json.loads(result.text)["results"]


from kontiguity.utils.imports import *

class ScriptExecutorScheduler(threading.Thread):
    def __init__(self, input_queue, script, output_queues = [], **args):
        super(ScriptExecutorScheduler, self).__init__(**args)
        self._input_queue = input_queue
        self._output_queues = output_queues
        self._script = script
        self.start()

    def run(self):
        script_executor = ScriptExecutor(self._script)
        while True:
            try:
                value = self._input_queue.get(timeout= 10)
            except Empty:
                break
            if value == "DONE":
                break
            args, sbatch = value
            output = script_executor.launch_script(args, sbatch=sbatch)
            for queue in self._output_queues:
                queue.put(output)

class ScriptExecutor():
    def __init__(self, script):
        self._script = script

    def launch_script(self, args_list, sbatch = False):
        if sbatch:
            subprocess.run(["sbatch", self._script] + args_list)
        else:
            subprocess.run(["bash", self._script] + args_list)


import click

from . import (
    load,
    retrieve,
    map,
    classify,
    pipeline
)

@click.group()
@click.pass_context
def cli(ctx):
    pass

cli.add_command(load.load)
cli.add_command(retrieve.retrieve)
cli.add_command(map.map)
cli.add_command(classify.classify)
cli.add_command(pipeline.pipeline)

if __name__ == "__main__":
    cli()


from kontiguity.utils.functions import *
from kontiguity.workers import ScriptExecutor as scexec

def create_table(name, index, WGSs, min_size = 1000):
    """Creates the table with the elements."""
    table_els = [
        {
            "name":name,
            "index":index,
            "fastq1":wgs[0],
            "fastq2":wgs[1] if paired else "",
            "is_paired":paired,
            "min_size":min_size
        }
        for wgs, paired in WGSs
    ]

    return pd.DataFrame.from_dict(table_els)

def retrieve_contigs(retrieval_dict, outpath, logan = False, no_tmp = False, threads = 8, sbatch = False, **sbatch_params):
    """Launches contigs retrieval."""

    # building script
    header = get_header(sbtach = sbatch, outpath = outpath, **sbatch_params)
    path_to_scripts = "/".join(__file__.split('/')[:-1])
    logan_script = path_to_scripts + "/scripts/retrieve/logan.sh"
    build_contigs_script = path_to_scripts + "/scripts/retrieve/build_contigs.sh"
    filter_script = path_to_scripts + "/scripts/retrieve/filter.sh"
    scripts_outpath = f"{outpath}/scripts"
    build_arborescence(scripts_outpath)
    script = write_script(header, {
        "logan":(logan_script, 3),
        "build":(build_contigs_script, 10),
        "filter":(filter_script, 6),
        }, scripts_outpath, name = "retrieve_contigs")

    # initialisations
    contigs_queue = Queue()
    workers = [
        scexec.ScriptExecutorScheduler(contigs_queue, script) for _ in range(threads)
    ]
    
    # feading fastas to the input queue
    for ref in retrieval_dict:
        species = ref.replace(" ","_")
        table_data = retrieval_dict[ref]

        # creating outfolder
        outfolder = f'{outpath}/{species}/contigs'
        build_arborescence(outfolder)

        # parameters
        for k in range(len(retrieval_dict[ref])):
            # parameters
            row_data = table_data.iloc[k]
            local_outpath = f"{outfolder}/{row_data['contigs']}"
            build_arborescence(outpath)
            process_stmp = f"{row_data['contigs'].split('_')[-1]}"
            min_len = f"{row_data['min_size']}"

            # index is looked for in an eventual dataset previously loaded with the load command if a path is not provided
            index = row_data["index"]
            if not os.path.isfile(index + '.fna') and not os.path.isfile(index + '.fa'):
                index = f"{outpath}/{species}/dataset/genomes/{index}"

            # logan params
            to_logan = "true" if logan else "false"
            accession = row_data["fastq1"].split("/")[-1].split('.')[0].split('_')[0] # retrieving the SRA id in the fastq name if provided

            # building params
            to_build = "true"

            # fastq check
            ## if it is not a path to a file, will be looked for in the kontiguity arborescence. If still not found, will return the ref which will be considered as a SRA id for Logan.
            paired_fastq = f"{outpath}/{species}/dataset/wgs/{row_data['fastq1']}_1.fastq"
            single_fastq = f"{outpath}/{species}/dataset/wgs/{row_data['fastq1']}.fastq"

            if os.path.exists(paired_fastq):
                is_paired = "true"
                fastq_R1  = paired_fastq
                fastq_R2 = f"{outpath}/{species}/dataset/wgs/{row_data['fastq1']}_2.fastq"
                fastq = "."
            elif os.path.exists(single_fastq):
                is_paired = "false"
                fastq_R1  = "."
                fastq_R2 = "."
                fastq = single_fastq
            else:
                fastq_R1= row_data["fastq1"] if row_data["is_paired"] else "."
                fastq_R2= row_data["fastq2"] if row_data["is_paired"] else "."
                fastq = row_data["fastq1"] if not row_data["is_paired"] else "."
                is_paired = "true" if row_data["is_paired"] else "false"

            # filtering params
            to_filter = "true"

            # queuing
            contigs_queue.put(([
                to_logan,
                to_build,
                to_filter,
                local_outpath, # logan parameters
                accession,
                process_stmp,
                index, # build params
                local_outpath,
                fastq_R1,
                fastq_R2,
                fastq,
                is_paired,
                min_len,
                f"{threads}",
                process_stmp,
                f'{"true" if no_tmp else "false"}',
                index, # filter params
                local_outpath,
                min_len,
                f"{threads}",
                process_stmp,
                f'{"true" if no_tmp else "false"}',
            ], sbatch))

    # closing queue and workers
    for _ in range(len(workers)):
        contigs_queue.put("DONE")

    return workers

def retrieve(
    name = "",
    outpath = "",
    index = "",
    min_size = 1000,
    wgs = "",
    table = None,
    logan = False,
    no_tmp = False,
    threads =  8,
    sbatch = False,
    sbtach_partition = 'dedicated',
    sbtach_qos = 'fast',
    sbtach_mem = '40G',
    sbatch_ncpus = 30
):
    sbatch_params = {
        '--partition': sbtach_partition,
        '--qos': sbtach_qos,
        '--mem': sbtach_mem,
        '-c': sbatch_ncpus
    }

    outfolder = f"{outpath}/{name.replace(' ', '_')}"
    build_arborescence(outfolder)

    ## building data tables
    if table is not None:
        data = pd.read_csv(table)
    else:
        data = create_table(name, index, wgs, min_size = min_size)
        if data is None:
            print("Error: missing WGS or HIC input.")
            return
    data["contigs"] = [f"contigs_{k + 1}" for k in range(len(data))]
    data.to_csv(f"{outfolder}/contigs_data.csv", index=False)
    
    ## retrieving unique name subset
    subset_retrieval = {}
    is_single_name = True
    for subname in np.unique(data['name']):
        if is_single_name and name != subname:
            is_single_name = False
        subset_retrieval[subname] = data[data['name'] == subname]

    ## launching retrievers
    out_tmp = outpath if is_single_name else outfolder
    retrievers = retrieve_contigs(subset_retrieval, outpath = out_tmp, logan = logan, no_tmp = no_tmp, threads = threads, sbatch = sbatch, **sbatch_params)

    ### joining loaders
    join_workers(retrievers)


from kontiguity.utils.functions import *    

class FastaFormater():
    def __init__(self, sequence_types=['chromosome', 'organelle'], batch_size = 20):
        self._sequence_types = sequence_types
        self._batch_size = batch_size

    @staticmethod
    def parse_ENA_info(info_line):
        """Parses the ENA sequence info line from fasta file"""
        infos = info_line.replace('\n','').split(" ")
        id = infos[0][1:]
        leftover = " ".join(infos[1:]).split(',')
        description = leftover[0]
        sequence_type = leftover[1].replace(" ", "").split(":")
        return {
            'id':id,
            'description':description,
            'sequence_type':sequence_type[0],
            'sequence_name':sequence_type[1]
        }

    def format_fasta(self, fasta_path, outfolder):
        """Writes a new fasta file keeping only the sequences with the set sequence types. Writes the chromosomes infos in a tsv file."""
        chromosome_infos = []
        chrom_file = None
        chromosomes = []
        batched_lines = []
        fasta_name = fasta_path.split('/')[-1][:-len(".all_seqs.fa")]
        output_fasta = open(outfolder + f"/{fasta_name}.filtered.fa", "w")
        has_chrom_info = os.path.isfile(f"{outfolder}/{fasta_name}.chromosomes.csv") or os.path.isfile(f"{outfolder}/chromosomes.csv")

        if has_chrom_info:
            if os.path.isfile(f"{outfolder}/chromosomes.csv"):
                shutil.copyfile(f"{outfolder}/chromosomes.csv", f"{outfolder}/{fasta_name}.chromosomes.csv")
            chrom_file = pd.read_csv(f"{outfolder}/{fasta_name}.chromosomes.csv")
            chromosomes = np.unique(chrom_file['id'])

        with open(fasta_path, "r") as fasta:
            keep_current = False
            line = fasta.readline()
            while len(line) > 1:
                if line[0] == ">":
                    if not has_chrom_info:
                        infos = self.parse_ENA_info(line)
                        keep_current = infos['sequence_type'] in self._sequence_types
                        if keep_current:
                            chromosome_infos.append(infos)
                    else:
                        chromosome = line[1:].split(' ')[0]
                        keep_current = chromosome in chromosomes and list(chrom_file[chrom_file['id'] == chromosome]['sequence_type'])[0] in self._sequence_types

                if keep_current:
                    batched_lines.append(line)
                    if len(batched_lines) >= self._batch_size:
                        output_fasta.writelines(batched_lines)
                        batched_lines = []
                line = fasta.readline()

            if len(batched_lines) > 0:
                output_fasta.writelines(batched_lines)

        output_fasta.close()

        if not has_chrom_info:
            chromosome_infos_df = pd.DataFrame.from_dict(chromosome_infos)
            chromosome_infos_df.to_csv(outfolder + f"/{fasta_name}.chromosomes.csv", sep=',')

        return outfolder + f"/{fasta_name}.filtered.fa"

fasta = sys.argv[1]
sequence_types = sys.argv[2].split(',')
outfolder = sys.argv[3]

formater = FastaFormater(sequence_types = sequence_types)
formater.format_fasta(fasta, outfolder)


from kontiguity.utils.functions import *
from kontiguity.workers import ScriptExecutor as scexec
from kontiguity.workers.DToLScrapper import DToLScrapperScheduler
from kontiguity.workers.DToLFormater import DToLFormaterScheduler

def create_table(name, ref, WGSs, HICs, no_hic = False, no_wgs = False):
    """Builds a kontiguity dataset table. At least one wgs and hic must be provided."""
    table_dict = []
    if no_hic and not no_wgs:
        for wgs in WGSs:
            table_dict.append({
                "name":name,
                "ref":ref,
                "wgs":wgs,
                "hic":"."
            })
    elif no_wgs and not no_hic:
        for hic in HICs:
            table_dict.append({
                "name":name,
                "ref":ref,
                "wgs":".",
                "hic":hic
            })
    elif no_wgs and no_hic:
        table_dict.append({
            "name":name,
            "ref":ref,
            "wgs":".",
            "hic":"."
        })
    else:
        for wgs in WGSs:
            for hic in HICs:
                table_dict.append({
                    "name":name,
                    "ref":ref,
                    "wgs":wgs,
                    "hic":hic
                })
    if len(table_dict) == 0:
        return None
    return pd.DataFrame.from_dict(table_dict)

def load_dtol(nb_per_page=100, threads=8):
    """Retrieves datas from the Darwin Tree of Life and builds a kontiguity dataset table from it."""

    # retrieving the count value
    URL_tmp = f"https://portal.darwintreeoflife.org/api/data_portal?limit=1&offset=0&sort=currentStatus:asc&current_class=kingdom"

    result = None
    try:
        result = requests.get(URL_tmp)
    except:
        print("Request failed")
    data = json.loads(result.text)
    count = data["count"]

    # initialisation
    offset = 0
    request_params_queue = Queue()
    result_queue = Queue()
    table_queue = Queue()

    # initialazing workers
    scrappers = [
        DToLScrapperScheduler(request_params_queue, result_queue)
        for _ in range(threads)
    ]
    formaters = [
        DToLFormaterScheduler(result_queue, table_queue)
        for _ in range(threads)
    ]

    # feeding input queue
    while offset <= count:
        request_params_queue.put((nb_per_page, offset))
        offset += nb_per_page
        
    # closing & joining
    for _ in range(threads):
        request_params_queue.put("DONE")
    for scrapper in scrappers:
        scrapper.join()
    for _ in range(threads):
        result_queue.put("DONE")
    for formater in formaters:
        formater.join()
    table_queue.put("DONE")

    # writing formated table
    table = []
    while True:
        try:
            res = table_queue.get(timeout=10)
        except Empty:
            break
        if res == "DONE":
            break
        table.append(res)

    species_df = pd.DataFrame.from_dict(table)    
    return species_df

def load_ref(ref_dict, outpath, chroms = None, threads = 8, sequence_types = 'chromosome,organelle', sbatch = False, **sbatch_params):
    """Builds the arborescence and eventualy retrieve fastas from the GCA database before building bowtie indexes."""
    # building script
    header = get_header(sbtach = sbatch, outpath = outpath, **sbatch_params)
    path_to_scripts = "/".join(__file__.split('/')[:-1])
    load_ref_script = path_to_scripts + "/scripts/loaders/load_ref.sh"
    format_ref_script = path_to_scripts + "/scripts/loaders/format_ref.sh"
    scripts_outpath = f"{outpath}/scripts"
    build_arborescence(scripts_outpath)
    script = write_script(header, {"load":(load_ref_script, 3), "format":(format_ref_script, 4)}, scripts_outpath, name = "load_ref")

    # initialisations
    genome_queue = Queue()
    workers = [
        scexec.ScriptExecutorScheduler(genome_queue, script) for _ in range(threads)
    ]

    # feading fastas to the input queue
    for ref in ref_dict:
        species = ref.replace(" ","_")

        # creating outfolder
        outfolder = f'{outpath}/{species}/dataset/genomes'
        build_arborescence(outfolder)

        # adding chromosome information to the outfolder if provided
        if chroms is not None and os.path.isfile(chroms):
            shutil.copyfile(chroms, f"{outfolder}/chromosomes.csv")

        # parameters
        for k in range(len(ref_dict[ref])):
            # scripts to call
            to_load = "true" if not os.path.isfile(ref_dict[ref][k]) else 'none'
            to_format = "true"

            # parameters
            fasta_path = f"https://www.ebi.ac.uk/ena/browser/api/fasta/{ref_dict[ref][k]}?download=true&gzip=true" if not os.path.exists(ref_dict[ref][k]) else ref_dict[ref][k]
            genome_name = species + f"_{k + 1}"
            loaded_fasta_path = f"{outfolder}/{genome_name}.all_seqs.fa" if not os.path.exists(ref_dict[ref][k]) else ref_dict[ref][k]

            # queuing
            genome_queue.put(([
                to_load,
                to_format,
                fasta_path,
                outfolder,
                genome_name,
                loaded_fasta_path,
                outfolder,
                genome_name,
                sequence_types
            ], sbatch))

    # closing queue and workers
    for _ in range(len(workers)):
        genome_queue.put("DONE")

    return workers

def load_fastqs(fastq_dict, outpath, experiment, threads = 8, sbatch = False, **sbatch_params):
    """Builds the arborescence and eventualy retrieve fastq files with fasterq-dump before splitting in subfastqs."""
    # building script
    header = get_header(sbtach = sbatch, outpath = outpath, **sbatch_params)
    path_to_scripts = "/".join(__file__.split('/')[:-1])
    load_fastq_script = path_to_scripts + "/scripts/loaders/load_fastq.sh"
    scripts_outpath = f"{outpath}/scripts"
    build_arborescence(scripts_outpath)
    script = write_script(header, {"load":(load_fastq_script, 3)}, scripts_outpath, name = "load_fastq")

    # initialisations
    fastq_queue = Queue()
    workers = [
        scexec.ScriptExecutorScheduler(fastq_queue, script) for _ in range(threads)
    ]

    # feading fastas to the input queue
    for ref in fastq_dict:
        species = ref.replace(" ","_")

        # creating outfolder
        outfolder = f'{outpath}/{species}/dataset/{experiment}'
        build_arborescence(outfolder)

        # parameters
        for k in range(len(fastq_dict[ref])):
            if not isinstance(fastq_dict[ref][k], str):
                fastqs, paired = fastq_dict[ref][k]
                fastq1 = fastqs[0]
                fastq2 = fastqs[1] if len(fastqs) > 1 else ""
            else:
                fastq1 = fastq_dict[ref][k]
                fastq2 = ""
                paired = False

            # scripts to call
            to_load1 = "true" if not os.path.isfile(fastq1) else 'none'
            to_load2 = "true" if paired == "paired" and not os.path.isfile(fastq2) else 'none'

            # queuing first fastq
            fastq_queue.put(([
                to_load1,
                fastq1,
                outfolder,
                str(threads)
            ], sbatch))

            # if paired, queuing second fastq
            if paired == "paired":
                fastq_queue.put(([
                    to_load2,
                    fastq2,
                    outfolder,
                    str(threads)
                ], sbatch))

    return workers

def load(
    name = "",
    outpath = None,
    ref = None,
    chroms = None,
    wgs = [],
    hic = [],
    table = None,
    dtol = False,
    no_wgs = False,
    no_hic = False,
    threads =  8,
    sbatch = False,
    sbtach_partition = 'dedicated',
    sbtach_qos = 'fast',
    sbtach_mem = '40G',
    sbatch_ncpus = 30
):
    sbatch_params = {
        '--partition': sbtach_partition,
        '--qos': sbtach_qos,
        '--mem': sbtach_mem,
        '-c': sbatch_ncpus
    }

    outfolder = f"{outpath}/{name.replace(' ', '_')}"
    build_arborescence(outfolder)

    ## building data tables
    if dtol:
        data = load_dtol(threads = threads)
    elif table is not None:
        data = pd.read_csv(table)
    else:
        data = create_table(name, ref, wgs, hic, no_hic = no_hic, no_wgs = no_wgs)
        if data is None:
            print("Error: missing WGS or HIC input.")
            return
    data.to_csv(f"{outfolder}/samples_data.csv", index=False)

    ## retrieving unique name subset
    subset_ref = {}
    subset_wgs = {}
    subset_hic = {}
    is_single_name = True
    for subname in np.unique(data['name']):
        if is_single_name and name != subname:
            is_single_name = False
        subset = data[data['name'] == subname]
        subset_ref[subname] = np.unique(subset['ref'])
        subset_wgs[subname] = np.unique(subset['wgs'])
        subset_hic[subname] = np.unique(subset['hic'])

    ## launching loaders
    outtmp = outpath if is_single_name else outfolder
    ref_loaders = load_ref(subset_ref, outpath = outtmp, chroms = chroms, threads = threads, sbatch = sbatch, **sbatch_params)
    wgs_loaders = load_fastqs(subset_wgs, outpath = outtmp, experiment='wgs', threads = threads, sbatch = sbatch, **sbatch_params) if not no_wgs else []
    hic_loaders = load_fastqs(subset_hic, outpath = outtmp, experiment='hic', threads = threads, sbatch = sbatch, **sbatch_params) if not no_hic else []

    ### joining loaders
    join_workers(ref_loaders)
    join_workers(wgs_loaders)
    join_workers(hic_loaders)


import click

from .types import *

import kontiguity.map as kmap

@click.command("map")
@click.option(
    '-n',
    '--name', 
    type=str,
    help="name of the experiment (recommanded: species name. info: spaces are not allowed and will be replaced by _.)"
)
@click.option(
    '-o',
    '--outpath',
    type=str,
    help="output folder path, created if non-existent"
)
@click.option(
    "-i",
    "--index",
    type=str,
    help="path to the reference genome index."
)
@click.option(
    "--hic",
    type=PAIR_LIST,
    help="path to the Hi-C fastq(s). If paired, provide both fastqs comma-separated."
)
@click.option(
    "--enzymes",
    type=str,
    default="DpnII,HinfI",
    help='comma-separated list of Hi-C restriction enzymes (dflt: DpnII,HinfI). The default enzymes where chosen in regard of the Arima Hi-C kit.'
)
@click.option(
    "--binnings",
    type=str,
    default="10000",
    help='comma separated bin sizes in bp in which each map is generated (dflt: 10000).'
)
@click.option(
    "--format",
    type=click.Choice(["cool", "mcool"]),
    default="cool",
    help='output format of generated maps in each binning: [cool/mcool] (dflt: cool).'
)
@click.option(
    "--table",
    type=str,
    help='path to a csv table providing the data parameters (Mandatory column heads: ["name", "index", "hic", "enzymes", "binnings"]).'
)
@click.option(
    "-t",
    "--threads",
    type=int,
    default=8,
    help='number of threads to launch for each subtask (dflt: 8)'
)
@click.option(
    "--no_tmp",
    is_flag=True,
    default=False,
    help="if selected, all the temporary files will be discarded. (dflt: False)"
)
@click.option(
    "--sbatch",
    is_flag=True,
    default=False,
    help="if selected, all the bash script will be launched as individual jobs on a SLURM distribution."
)
@click.option(
    "--sbtach_partition",
    default='dedicated',
    type=str,
    help="partition requested for sbatch."
)
@click.option(
    "--sbtach_qos",
    default= 'fast',
    type=str,
    help="quality of service required for sbatch."
)
@click.option(
    "--sbtach_mem",
    default='40G',
    type=str,
    help="minimum amount of real memory requested for sbatch."
)
@click.option(
    "--sbatch_ncpus",
    default=30,
    type=int,
    help="number of cpus required per task fro sbatch."
)
def map(**args):
    kmap.map(**args)


from .imports import *

def build_arborescence(path):
    """Creates the necessary folders for the provided path to be a valid folder path."""
    folders = path.split('/')
    
    for i in range(1, len(folders) + 1):
        current_path = "/".join(folders[:i])
        if not os.path.exists(current_path) and current_path != "":
            os.mkdir(current_path)

def join_workers(workers):
    """Joins each worker in the workers list."""
    for worker in workers:
        worker.join()

def get_header(sbtach = False, outpath = None, **sbatch_params):
    """Gets the header of the bash script to execute. Applies sbatch parameters if provided."""
    header = "#!/bin/bash\n"
    if sbtach:
        # adding sbatch defined parameters
        for param in sbatch_params:
            header += f"#SBATCH {param}={sbatch_params[param]}\n"

        # adding output path
        header += f"#SBATCH -o {outpath}/logs/R-%x-%j-reference_load.out.txt -e {outpath}/logs/R-%x-%j-reference_load.err.txt\n"
    return header

def write_script(header, subscripts, outpath, name = "script"):
    """Writes a script that executes the subscripts with condition (true, false). Parameters will be passed in the order of the scripts."""
    ### when calling the script, ALL the parameters must be provided, even for scripts set to false (none will do perfectly).
    ### script first arguments are the execution condition of each script, in order (set to "true" to execute, anything else not to).
    script_path = outpath + f"/{name}.sh"
    with open(script_path, 'w') as script:
        script.write(header + "\n")
        nb_args = 1 + len(subscripts)
        nb_scripts = 1
        for script_name in subscripts:
            subscript, args = subscripts[script_name]
            script.write(f'if [ ${nb_scripts} = "true" ];then\n')
            nb_scripts += 1
            arguments = ""
            for arg in range(1,args + 1):
                script.write(f"\tparam_{arg}=$" + "{" + f"{nb_args}" + "}\n")
                arguments += f"\t$param_{arg} "
                nb_args += 1
            script.write(f"\tbash {subscript} {arguments}\n")
            script.write('fi\n\n')

    return script_path



import click

import kontiguity.classify as kclassify

@click.command("classify")
@click.option(
    '-n',
    '--name', 
    type=str,
    help="name of the experiment (recommanded: species name. info: spaces are not allowed and will be replaced by _.)"
)
@click.option(
    '-o',
    '--outpath',
    type=str,
    help="output folder path, created if non-existent"
)
@click.option(
    "--chroms",
    type=str,
    help='path to a chromosome information file detailing the type of each sequence present in the reference (Mandatory column heads: ["id", "sequence_type", "sequence_name"]). "sequence_type" must be in the ENA database format : ["chromosome", "organelle", ...]. Required only for a local fasta, GCA referenced genomes will have the chromosome.tsv generated.'
)
@click.option(
    "--mcool",
    type=str,
    help="path to mcool file of contigs to classify. The program will compute and classify the contact profiles of contigs not referenced in the chromosome info file. Requires --binning."
)
@click.option(
    "--binnings",
    type=str,
    default="10000",
    help='comma separated bin sizes in bp in which each map is generated (dflt: 10000).'
)
@click.option(
    "--cool",
    type=str,
    help="ppath to cool file of contigs to classify. The program will compute and classify the contact profiles of contigs not referenced in the chromosome info file."
)
@click.option(
    "--model",
    type=str,
    default="models/model.hdf5",
    help="path to a classifier in hdf5 format (dflt: provided model)."
)
@click.option(
    "--param_file",
    type=str,
    default="models/model.json",
    help="path to a json file containing the data preprodessing and model parameters (dflt: model params file)"
)
@click.option(
    "--table",
    type=str,
    help='path to a csv table providing the data parameters (Mandatory column heads: ["name", "ref", "wgs", "hic"]).'
)
def classify(**args):
    click.echo(args)
    kclassify.classify(**args)


from kontiguity.utils.functions import *
from kontiguity.workers import ScriptExecutor as scexec

def create_table(name, index, HICs, enzymes = "HinfI,DpnII", binnings = "5000", format = "cool"):
    """Creates the table with the elements."""
    table_els = [
        {
            "name":name,
            "index":index,
            "fastq1":hic[0],
            "fastq2":hic[1] if paired else "",
            "enzymes":enzymes,
            "binnings":binnings,
            "format":format
        }
        for hic, paired in HICs
    ]

    return pd.DataFrame.from_dict(table_els)

def map_hics(hic_dict, outpath, no_tmp = False, threads = 8, sbatch = False, **sbatch_params):
    """Launches Hi-C mapping."""

    # building script
    header = get_header(sbtach = sbatch, outpath = outpath, **sbatch_params)

    path_to_scripts = "/".join(__file__.split('/')[:-1])
    generate_cool_script = path_to_scripts + "/scripts/map/generate_cool.sh"
    rebin_script = path_to_scripts + "/scripts/map/rebin.sh"
    zoomify_script = path_to_scripts + "/scripts/map/zoomify.sh"

    scripts_outpath = f"{outpath}/scripts"
    build_arborescence(scripts_outpath)

    script = write_script(header, {
        "generate":(generate_cool_script, 8),
        "rebin":(rebin_script, 6),
        "zoomify":(zoomify_script, 6),
        }, scripts_outpath, name = "map_hic")
    
    # initialisations
    hic_queue = Queue()
    workers = [
        scexec.ScriptExecutorScheduler(hic_queue, script) for _ in range(threads)
    ]
    
    # feading fastas to the input queue
    for ref in hic_dict:
        species = ref.replace(" ","_")
        table_data = hic_dict[ref]

        # creating outfolder
        outfolder = f'{outpath}/{species}/maps'
        build_arborescence(outfolder)

        # parameters
        for k in range(len(table_data)):
            # parameters
            row_data = table_data.iloc[k]
            local_outpath = f"{outfolder}/{row_data['mapping']}"
            build_arborescence(outpath)
            hic_name = row_data['mapping']
            binnings = sorted(list(row_data['binnings'].split(',')))
            min_binning = str(reduce(math.gcd, [int(value) for value in binnings]))
            other_binnings = binnings[1:] if min_binning in binnings else binnings

            # index is looked for in an eventual dataset previously loaded with the load command if a path is not provided
            index = row_data["index"]
            if not os.path.isfile(index + '.fna') and not os.path.isfile(index + '.fa'):
                index = f"{outpath}/{species}/dataset/genomes/{index}"

            # generate map params
            to_map = "true"
            enzymes=row_data['enzymes']

            # fastq check
            ## if it is not a path to a file, will be looked for in the kontiguity arborescence.
            paired_fastq = f"{outpath}/{species}/dataset/hic/{row_data['fastq1']}_1.fastq"
            if os.path.exists(paired_fastq):
                hic_R1  = paired_fastq
                hic_R2 = f"{outpath}/{species}/dataset/hic/{row_data['fastq1']}_2.fastq"
            else:
                hic_R1= "." if len(row_data["fastq1"]) == 0 else row_data["fastq1"]
                hic_R2= "." if len(row_data["fastq2"]) == 0 else row_data["fastq2"]

            # rebin params
            to_rebin = "true" if row_data['format'] == "cool" and len(other_binnings) > 0 else "false"
            binnings_formated = " ".join(other_binnings)

            # zoomify params
            to_zoomify = "true" if row_data['format'] == "mcool" else "false"
            zoomings = ",".join(binnings)

            # queuing
            hic_queue.put(([
                to_map,
                to_rebin,
                to_zoomify,
                index, # generate map
                local_outpath,
                hic_name,
                hic_R1,
                hic_R2,
                enzymes,
                min_binning,
                str(threads),
                local_outpath, # rebin
                hic_name,
                min_binning,
                binnings_formated,
                str(threads),
                "true" if no_tmp else "false",
                local_outpath, # zoomify
                hic_name,
                min_binning,
                zoomings,
                str(threads),
                "true" if no_tmp else "false"
            ], sbatch))

    # closing queue and workers
    for _ in range(len(workers)):
        hic_queue.put("DONE")

    return workers

def map(
    name = "",
    outpath = "",
    index = "",
    hic = "",
    enzymes = "HinfI,DpnII",
    binnings = "5000",
    format = "cool",
    table = None,
    no_tmp = False,
    threads =  8,
    sbatch = False,
    sbtach_partition = 'dedicated',
    sbtach_qos = 'fast',
    sbtach_mem = '40G',
    sbatch_ncpus = 30
):
    sbatch_params = {
        '--partition': sbtach_partition,
        '--qos': sbtach_qos,
        '--mem': sbtach_mem,
        '-c': sbatch_ncpus
    }

    outfolder = f"{outpath}/{name.replace(' ', '_')}"
    build_arborescence(outfolder)

    ## building data tables
    if table is not None:
        data = pd.read_csv(table)
    else:
        data = create_table(name, index, hic, enzymes = enzymes, binnings = binnings, format = format)
        if data is None:
            print("Error: missing HIC input.")
            return
    data["mapping"] = [f"mapping_{k + 1}" for k in range(len(data))]
    data.to_csv(f"{outfolder}/mapping_data.csv", index=False)
    
    ## retrieving unique name subset
    subset_hic = {}
    is_single_name = True
    for subname in np.unique(data['name']):
        if is_single_name and name != subname:
            is_single_name = False
        subset_hic[subname] = data[data['name'] == subname]

    ## launching retrievers
    out_tmp = outpath if is_single_name else outfolder
    retrievers = map_hics(subset_hic, outpath = out_tmp, no_tmp = no_tmp, threads = threads, sbatch = sbatch, **sbatch_params)

    ### joining loaders
    join_workers(retrievers)


import click

# import cooler
# import os

# # accepts a list of cool files or a file containing a list of cool, one per line
# class coolType(click.ParamType):
#     name="cool"
# # accepts a list of cool files or a file containing a list of cool, one per line
# class coolType(click.ParamType):
#     name="cool"

#     def is_cool(self, file):
#         try:
#             return cooler.fileops.is_cooler(file)
#         except:
#             return False
#     def is_cool(self, file):
#         try:
#             return cooler.fileops.is_cooler(file)
#         except:
#             return False
        
#     def is_mcool(self, file):
#         try:
#             return cooler.fileops.is_multires_file(file)
#         except:
#             return False
#     def is_mcool(self, file):
#         try:
#             return cooler.fileops.is_multires_file(file)
#         except:
#             return False

#     def convert(self, value, param, ctx):
#         if self.is_cool(value) or self.is_mcool(value):
#             return [value]
#         if not os.path.isfile(value):
#             files = value.split(',')
#             for file in files:
#                 if not self.is_cool(file) and not self.is_mcool(file):
#                     self.fail(f"{file} is not a valid cool file", param, ctx)
#             return files
#         else:
#             files = []
#             with open(value, 'r') as f:
#                 for line in f.readlines():
#                     file = line.replace('\n', '').replace(' ', '')
#                     if not self.is_cool(file) and not self.is_mcool(file):
#                         self.fail(f"{file} is not a valid cool file", param, ctx)
#                     else:
#                         files.append(file)
#             return files
#     def convert(self, value, param, ctx):
#         if self.is_cool(value) or self.is_mcool(value):
#             return [value]
#         if not os.path.isfile(value):
#             files = value.split(',')
#             for file in files:
#                 if not self.is_cool(file) and not self.is_mcool(file):
#                     self.fail(f"{file} is not a valid cool file", param, ctx)
#             return files
#         else:
#             files = []
#             with open(value, 'r') as f:
#                 for line in f.readlines():
#                     file = line.replace('\n', '').replace(' ', '')
#                     if not self.is_cool(file) and not self.is_mcool(file):
#                         self.fail(f"{file} is not a valid cool file", param, ctx)
#                     else:
#                         files.append(file)
#             return files

# class IntListType(click.ParamType):
#     name="int_list"
# class IntListType(click.ParamType):
#     name="int_list"

#     def convert(self, value, param, ctx):
#         list = value.split(',')
#         ints = []
#         for i in list:
#             if not isinstance(int(i), int):
#                 self.fail(f"{value} is not a comma-separated integer list.")
#             if int(i) < 0:
#                 self.fail(f"All integers provided in {param.name} must be positive integers.")
#             ints.append(int(i))
#         return ints
#     def convert(self, value, param, ctx):
#         list = value.split(',')
#         ints = []
#         for i in list:
#             if not isinstance(int(i), int):
#                 self.fail(f"{value} is not a comma-separated integer list.")
#             if int(i) < 0:
#                 self.fail(f"All integers provided in {param.name} must be positive integers.")
#             ints.append(int(i))
#         return ints
    
# class StrListType(click.ParamType):
#     name="str_list"

#     def convert(self, value, param, ctx):
#         strs = value.split(',')
#         for s in strs:
#             if len(s) == 0:
#                 self.fail(f"Empty strings. Comma must separate two distinct values; if a single string is passed, no comma must be found in {param.name}.")
#         return strs
    
class PairListType(click.ParamType):
    name="pair_list"

    def convert(self, value, param, ctx):
        pairs = value.split(',')
        formated_pairs = []
        for pair in pairs:
            if len(pair) == 0:
                self.fail(f"Empty strings. Comma must separate two distinct values; if a single string is passed, no comma must be found in {param.name}.")
            pair_values = pair.split(':')
            if len(pair_values) > 2:
                self.fail(f"Too many values provided in {pair}. Colons must separate at most two distinct values; if a single value is passed, do not add colons.")
            is_paired = len(pair_values) == 2
            formated_pairs.append((pair_values, is_paired))
        return formated_pairs

# COOL = coolType()
# INT_LIST = IntListType()
# STR_LIST = StrListType()
PAIR_LIST = PairListType()


import click

import kontiguity.pipeline as kpipeline

@click.command("pipeline")
@click.option(
    '-n',
    '--name', 
    type=str,
    help="name of the experiment (recommanded: species name. info: spaces are not allowed and will be replaced by _.)"
)
@click.option(
    '-o',
    '--outpath',
    type=str,
    help="output folder path, created if non-existent"
)
@click.option(
    "-r",
    "--ref",
    type=str,
    help="path to the reference genome fasta OR the GCA reference which will automatically be loaded from ENA database."
)
@click.option(
    "--chroms",
    type=str,
    help='path to a chromosome information file detailing the type of each sequence present in the reference (Mandatory column heads: ["id", "sequence_type", "sequence_name"]). "sequence_type" must be in the ENA database format : ["chromosome", "organelle", ...]. Required only for a local fasta, GCA referenced genomes will have the chromosome.tsv generated.'
)
@click.option(
    "--wgs",
    type=str,
    help="path to the WGS fastq(s) OR SRA accession. If paired and local, provide both fastqs comma-separated."
)
@click.option(
    "--hic",
    type=str,
    help="path to the Hi-C fastq(s) OR SRA accession. If paired and local, provide both fastqs comma-separated."
)
@click.option(
    "--table",
    type=str,
    help='path to a csv table providing the data parameters (Mandatory column heads: ["name", "ref", "wgs", "hic"]).'
)
@click.option(
    "--start",
    type=str,
    help='starting point of the pipeline. If some processes have been started at this step, the pipeline will pick-up where it left, except in the --no-pickup mode. Requires the arboresence of any previous step to exist. To select in ["load", "retrieve", "map", "classify"] (dflt: load).'
)
@click.option(
    "--stop",
    type=str,
    help='stopping point of the pipeline. To select in ["load", "retrieve", "map", "classify"] (dflt: classify).'
)
@click.option(
    "--no-pickup",
    is_flag=True,
    default=False,
    help="if selected, will restart avery process of the selected starting point in the pipeline."
)
def pipeline(**args):
    click.echo(args)
    kpipeline.pipeline(**args)


from kontiguity.utils.imports import *

class DToLFormaterScheduler(threading.Thread):
    def __init__(self, input_queue, output_queue, **args):
        super(DToLFormaterScheduler, self).__init__(**args)
        self._input_queue = input_queue
        self._output_queue = output_queue
        self.start()

    def run(self):
        formater = DToLFormater()
        while True:
            try:
                value = self._input_queue.get()
            except Empty:
                break
            if value == "DONE":
                break
            output = formater.extract_infos(value)
            if output is not None:
                species, genomes, HiCs, WGSs = output
                for row in formater.format(species, genomes, HiCs, WGSs):
                    self._output_queue.put(row)

class DToLFormater():
    def __init__(self):
        pass

    @staticmethod
    def format(species, genomes, HiCs, WGSs):
        """creates a table row for each combination of assembly, Hi-C and WGS for the species."""
        for i in range(len(genomes)):
            for j in range(len(WGSs)):
                for k in range(len(HiCs)):
                    yield {
                        "species": species,
                        "name": species,
                        **genomes[i],
                        **WGSs[j],
                        **HiCs[k]
                    }
    
    @staticmethod
    def extract_infos(species_data):
        species = species_data["_id"]

        # retrieving Hi-C and WGS. If no Hi-C or WGS is available, returns None
        HiCs = []
        WGSs = []
        column = {
            "Hi-C":"hic",
            "WGS":"wgs"
        }
        for experiment in species_data["_source"]["experiment"]:
            strategy = experiment["library_strategy"]
            if strategy in ["Hi-C", "WGS"]:
                exp_datas = {
                    f"{column[strategy]}": experiment["study_accession"],
                    f"protocol_{strategy}": experiment["library_construction_protocol"] if len(experiment["library_construction_protocol"]) > 0 else strategy,
                    f"run_acession_{strategy}": experiment["run_accession"],
                    f"layout_{strategy}": experiment["library_layout"]
                }
            match strategy:
                case "Hi-C":
                    HiCs.append(exp_datas)
                case "WGS":
                    WGSs.append(exp_datas)
        
        if len(HiCs) == 0 or len(WGSs) == 0:
            return None

        genomes = []
        if "annotation" in species_data["_source"]: #  if the species is annotated, the reference genome used will be the one corresponding to the annotation. 
            for annotation in species_data["_source"]["annotation"]:
                genomes.append({
                    "ref": annotation["accession"],
                    # "study_accession":"",
                    # "fasta":f"https://www.ebi.ac.uk/ena/browser/api/fasta/{annotation["accession"]}?download=true&gzip=true",
                })


        # Useless: only retrieving 20 more species for a several times longer execution time
        # elif "assemblies" in species_data["_source"]: # Otherwise, we retrieve all the available genomes from the assemblies.
        #     for assembly in species_data["_source"]["assemblies"]:
        #         genome_location = f"https://ftp.ebi.ac.uk/pub/ensemblorganisms/{species.replace(' ', '_')}/{assembly['accession']}.{assembly['version']}/genome"
        #         # checking the address exists
        #         try:
        #             time.sleep(0.5) # sleeping to avoid max retries connection refused from EBI
        #             response = requests.get(genome_location)
        #             if response.status_code != 200:
        #                 continue
        #         except Exception as e:
        #             continue
        #         genomes.append({
        #             "accession": assembly["accession"],
        #             "study_accession": assembly["study_accession"],
        #             "fasta": f"{genome_location}/softmasked.fa.gz",
        #             "chromosomes": f"{genome_location}/chromosomes.tsv.gz",
        #         })

        if len(genomes) == 0:
            return None
        
        return species, genomes, HiCs, WGSs



# data managment
import pandas as pd
import numpy as np

# system
import subprocess
import datetime
import shutil
import sys
import os

# multitasking
from queue import Queue, Empty
import threading

# network
import requests
import json

# mathematical operations
import math
from functools import reduce


# filters fasta sequences per length and keeps only the ones above a given length in output_fasta
from kontiguity.utils.functions import *

def filter_fasta_per_length(fasta, length, output_fasta):

    output_folder = "/".join(output_fasta.split('/')[:-1])
    build_arborescence(output_folder)

    filtered = open(output_fasta, 'w')

    current_id = ""
    current_lines = []
    current_length = 0

    with open(fasta, 'r') as unfiltered:
        line = unfiltered.readline()
        while line is not None and len(line) > 0:
            if line[0] == '>':
                # writing current sequence
                if current_length >= length:
                    filtered.writelines([current_id] + current_lines)
                current_id = line
                current_lines = []
                current_length = 0
            else:
                current_lines.append(line)
                current_length += len(line.replace("\n", ""))

            line = unfiltered.readline()

    if current_length >= length:
        filtered.writelines([current_id] + current_lines)
    
    filtered.close()

if len(sys.argv) == 1: # local test
    local_path = "/".join(__file__.split('/')[:-5])
    fasta = local_path + "/test_data/retrieve/contigs.fa"
    min_length = 1000
    output_fasta = local_path + f"/test_results/filtering/contigs_filtered_{min_length}.fa"

else: # global usage
    fasta = sys.argv[1]
    min_length = int(sys.argv[2])
    output_fasta = sys.argv[3]

filter_fasta_per_length(fasta, min_length, output_fasta)

if len(sys.argv) == 1:
    assert(os.path.isfile(output_fasta))
